import subprocess
import sys
import tempfile
import textwrap
import warnings
from typing import Literal
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import requests
import soundfile as sf
from pydantic import ValidationError

from mistral_common.audio import hertz_to_mel, mel_filter_bank
from mistral_common.protocol.instruct.chunk import AudioChunk, RawAudio
from mistral_common.protocol.transcription.request import TranscriptionRequest
from mistral_common.tokens.tokenizers.audio import Audio, AudioConfig, AudioSpectrogramConfig


def _make_dummy_base64() -> str:
    """Create a minimal valid base64-encoded WAV for testing."""
    audio = Audio(audio_array=np.zeros(1600, dtype=np.float32), sampling_rate=16000, format="wav")
    return audio.to_base64("wav")


def sin_wave(sampling_rate: int, duration: float) -> np.ndarray:
    return np.sin(np.ones([int(duration * sampling_rate)]))


def test_audio_config_rejects_sampling_rate_below_frame_rate() -> None:
    encoding_config = AudioSpectrogramConfig(num_mel_bins=80, hop_length=160, window_size=400)
    with pytest.raises(AssertionError):
        AudioConfig(sampling_rate=1, frame_rate=2.0, encoding_config=encoding_config)


def test_audio_resample() -> None:
    sampling_rate = 44_000
    original_array = sin_wave(sampling_rate, 1)

    audio = Audio(
        audio_array=original_array,
        sampling_rate=sampling_rate,
        format="wav",
    )

    audio.resample(sampling_rate)
    assert np.allclose(audio.audio_array, original_array, atol=1e-5)

    audio.resample(sampling_rate // 2)
    assert audio.sampling_rate == sampling_rate // 2
    assert len(audio.audio_array) == len(original_array) // 2
    # fmt: off
    expected_resampled_array = [0.622, 0.907, 0.802, 0.870, 0.819, 0.860, 0.826, 0.854, 0.831, 0.850]
    # fmt: on
    assert np.allclose(audio.audio_array[:10], expected_resampled_array, atol=1e-3)


def test_from_file() -> None:
    sampling_rate = 44100
    original_array = sin_wave(sampling_rate, 1)

    # Test with a local path
    with tempfile.NamedTemporaryFile(suffix=".wav") as tmp:
        with sf.SoundFile(tmp.name, "w", samplerate=sampling_rate, channels=1) as f:
            f.write(original_array)
        audio_local = Audio.from_file(tmp.name)
    assert isinstance(audio_local, Audio)
    assert np.allclose(audio_local.audio_array, original_array, atol=1e-5)
    assert audio_local.sampling_rate == sampling_rate


def test_from_url() -> None:
    # Test with an invalid URL
    with patch("mistral_common.tokens.tokenizers.audio._requests_lib.get") as mock_get:
        mock_get.side_effect = requests.RequestException("connection failed")
        with pytest.raises(
            ValueError,
            match="Failed to download audio from URL: https://example.com/invalid_audio.wav",
        ):
            Audio.from_url("https://example.com/invalid_audio.wav")

    # Test with an invalid content
    with patch("mistral_common.tokens.tokenizers.audio._requests_lib.get") as mock_get:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b"<html>this is not audio</html>"
        mock_get.return_value = mock_response
        with pytest.raises(
            ValueError,
            match=r"Failed to create Audio instance from URL: https://example\.com \.",
        ):
            Audio.from_url("https://example.com")

    # Test valid URL
    sampling_rate = 16_000
    audio_array = sin_wave(sampling_rate, 1)
    with tempfile.NamedTemporaryFile(suffix=".wav") as tmp:
        with sf.SoundFile(tmp.name, "w", samplerate=sampling_rate, channels=1) as f:
            f.write(audio_array)
        with open(tmp.name, "rb") as f:
            wav_bytes = f.read()

    with patch("mistral_common.tokens.tokenizers.audio._requests_lib.get") as mock_get:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = wav_bytes
        mock_get.return_value = mock_response
        audio_url = Audio.from_url("https://fake-audio-url.com/sample.wav", strict=False)
        assert isinstance(audio_url, Audio)


@pytest.mark.parametrize("prefix", [True, False])
def test_audio_base64(prefix: bool) -> None:
    sampling_rate = 16_000
    original_array = sin_wave(sampling_rate, 3.3)

    LOSSY = ["mp3"]
    LOSSLESS = ["wav"]

    for format in LOSSY + LOSSLESS:
        audio = Audio(
            audio_array=original_array,
            sampling_rate=sampling_rate,
            format=format,
        )

        base64_str = audio.to_base64(format, prefix)
        new_audio = Audio.from_base64(base64_str)

        if prefix:
            assert base64_str.startswith(f"data:audio/{format};base64,")

        assert audio.sampling_rate == new_audio.sampling_rate
        if format in LOSSLESS:
            assert np.allclose(audio.audio_array, new_audio.audio_array, atol=1e-5)
        elif format in LOSSY:

            def rmse(a: np.ndarray, b: np.ndarray) -> float:
                return float(np.sqrt(np.mean((a - b) ** 2)))

            assert rmse(audio.audio_array, new_audio.audio_array) < 5e-3
        else:
            raise ValueError(f"Unknown format {format}")


@pytest.mark.parametrize(
    "freq, expected_mel",
    [
        (100.0, 1.5),  # Linear region
        (1000.0, 15.0),  # Boundary of linear and log regions
        (4000.0, 15.0 + np.log(4000.0 / 1000.0) * 27.0 / np.log(6.4)),  # Log region
        (10000.0, 15.0 + np.log(10000.0 / 1000.0) * 27.0 / np.log(6.4)),  # Log region
    ],
)
def test_hertz_to_mel_single_value(freq: float, expected_mel: float) -> None:
    mel = hertz_to_mel(freq)
    assert isinstance(mel, float)
    assert abs(mel - expected_mel) < 1e-6


def test_hertz_to_mel_array() -> None:
    # Test with an array of frequency values
    freq_array = np.array([100.0, 1000.0, 4000.0])
    expected_mel_array = np.array(
        [
            3.0 * 100.0 / 200.0,  # Linear region
            15.0,  # min_log_mel
            15.0 + np.log(4000.0 / 1000.0) * 27.0 / np.log(6.4),  # Log region
        ]
    )
    np.testing.assert_array_equal(hertz_to_mel(freq_array), expected_mel_array)


def test_mel_filter_bank() -> None:
    # Test mel_filter_bank function
    num_frequency_bins = 256
    num_mel_bins = 20
    min_frequency = 0
    max_frequency = 8000
    sampling_rate = 16000

    mel_filters = mel_filter_bank(num_frequency_bins, num_mel_bins, min_frequency, max_frequency, sampling_rate)

    # Check the shape of the output
    assert mel_filters.shape == (num_frequency_bins, num_mel_bins)

    # Check that all filters sum to 1 (approximately)
    assert np.allclose(mel_filters.sum(axis=0), 0.032, atol=0.005)

    # Check that there are no negative values
    assert (mel_filters >= 0).all()

    # integration test
    expected_array = np.array(
        [
            0.03160475,
            0.03200728,
            0.03184327,
            0.03176877,
            0.03208179,
            0.03160917,
            0.0320748,
            0.03180101,
            0.03186299,
            0.03190694,
            0.0319008,
            0.03182839,
            0.03191571,
            0.03184625,
            0.03188972,
            0.03185817,
            0.0318928,
            0.03186744,
            0.03187269,
            0.0318759,
        ]
    )
    diff = np.abs(mel_filters.sum(0) - expected_array)
    assert (diff < 1e-5).all()


class TestDeprecationWarnings:
    """Each test runs in a fresh subprocess to avoid module-level _warned flags."""

    @staticmethod
    def _run_code(code: str) -> subprocess.CompletedProcess[str]:
        return subprocess.run(
            [sys.executable, "-W", "all", "-c", textwrap.dedent(code)],
            capture_output=True,
            text=True,
            timeout=30,
        )

    def test_rawaudio_instantiation_warns_once(self) -> None:
        result = self._run_code("""\
            import warnings
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                from mistral_common.protocol.instruct.chunk import RawAudio
                _ = RawAudio(data="abc", format="wav")
                _ = RawAudio(data="xyz", format="wav")
            dep = [x for x in w if issubclass(x.category, DeprecationWarning) and "RawAudio" in str(x.message)]
            assert len(dep) == 1, f"Expected 1 DeprecationWarning, got {len(dep)}: {dep}"
            print("OK")
        """)
        assert result.returncode == 0, result.stderr
        assert "OK" in result.stdout

    def test_audio_from_raw_audio_warns(self) -> None:
        result = self._run_code("""\
            import warnings
            import numpy as np
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                from mistral_common.tokens.tokenizers.audio import Audio
                from mistral_common.protocol.instruct.chunk import RawAudio
                audio_obj = Audio(audio_array=np.zeros(1600, dtype=np.float32), sampling_rate=16000, format="wav")
                b64 = audio_obj.to_base64("wav")
                ra = RawAudio(data=b64, format="wav")
                result = Audio.from_raw_audio(ra)
            dep = [x for x in w if issubclass(x.category, DeprecationWarning) and "from_raw_audio" in str(x.message)]
            assert len(dep) == 1, f"Expected 1 from_raw_audio warning, got {len(dep)}: {dep}"
            assert hasattr(result, "audio_array")
            assert result.sampling_rate > 0
            print("OK")
        """)
        assert result.returncode == 0, result.stderr
        assert "OK" in result.stdout

    @pytest.mark.parametrize("deprecated_name", ["Audio", "AudioFormat", "EXPECTED_FORMAT_VALUES"])
    def test_audio_import_from_old_location_warns(
        self, deprecated_name: Literal["Audio", "AudioFormat", "EXPECTED_FORMAT_VALUES"]
    ) -> None:
        code = (
            "import warnings\n"
            "import numpy as np\n"
            "with warnings.catch_warnings(record=True) as w:\n"
            '    warnings.simplefilter("always")\n'
            f"    from mistral_common.audio import {deprecated_name}\n"
            "dep = [\n"
            "    x for x in w\n"
            "    if issubclass(x.category, DeprecationWarning)\n"
            '    and "mistral_common.audio" in str(x.message)\n'
            "]\n"
            'assert len(dep) == 1, f"Expected 1 import warning, got {len(dep)}: {dep}"\n'
            "# Verify the class is usable\n"
        )
        if deprecated_name == "Audio":
            code += (
                'a = Audio(audio_array=np.zeros(1600, dtype=np.float32), sampling_rate=16000, format="wav")\n'
                "assert a.sampling_rate == 16000\n"
            )
        elif deprecated_name == "AudioFormat":
            code += "assert isinstance(AudioFormat.MP3, AudioFormat)\n"
        else:
            code += 'assert "mp3" in EXPECTED_FORMAT_VALUES\n'
        code += 'print("OK")\n'

        result = self._run_code(code)
        assert result.returncode == 0, result.stderr
        assert "OK" in result.stdout

    def test_rawaudio_import_from_chunk_still_works(self) -> None:
        result = self._run_code("""\
            import warnings
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                from mistral_common.protocol.instruct.chunk import RawAudio
                ra_str = RawAudio(data="hello", format="wav")
                assert ra_str.data == "hello"
                ra_bytes = RawAudio(data=b"world", format="wav")
                assert ra_bytes.data == b"world"
            print("OK")
        """)
        assert result.returncode == 0, result.stderr
        assert "OK" in result.stdout


class TestRawAudioBackwardCompat:
    def test_rawaudio_still_works(self) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            ra = RawAudio(data="some_data", format="wav")
        assert ra.data == "some_data"

    def test_rawaudio_format_accepted_silently(self) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            ra = RawAudio(data="data", format="wav")
        assert ra.format == "wav"

    def test_audiochunk_with_rawaudio_backward_compat(self) -> None:
        b64 = _make_dummy_base64()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            ra = RawAudio(data=b64, format="wav")
            chunk = AudioChunk(input_audio=ra)  # type: ignore[arg-type]
        assert chunk.input_audio == b64

    def test_audiochunk_with_rawaudio_dict_backward_compat(self) -> None:
        b64 = _make_dummy_base64()
        chunk = AudioChunk.model_validate({"input_audio": {"data": b64, "format": "wav"}})
        assert chunk.input_audio == b64

    def test_audiochunk_with_str_new_api(self) -> None:
        chunk = AudioChunk(input_audio="base64data")
        assert chunk.input_audio == "base64data"

    def test_audiochunk_with_bytes_new_api(self) -> None:
        chunk = AudioChunk(input_audio=b"raw_bytes")
        assert chunk.input_audio == b"raw_bytes"

    @pytest.mark.parametrize("empty_audio", ["", "   ", b""])
    def test_audiochunk_empty_input_raises_validation_error(self, empty_audio: str | bytes) -> None:
        with pytest.raises(ValidationError, match="should not be empty"):
            AudioChunk(input_audio=empty_audio)


def test_transcription_request_with_rawaudio_backward_compat() -> None:
    b64 = _make_dummy_base64()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        ra = RawAudio(data=b64, format="wav")
        req = TranscriptionRequest(audio=ra, language=None, target_streaming_delay_ms=None)  # type: ignore[arg-type]
    assert isinstance(req.audio, str)
    assert req.audio == b64



import json

import pytest

from mistral_common.protocol.instruct.chunk import (
    AudioChunk,
    ChunkTypes,
    ImageURLChunk,
    TextChunk,
    ThinkChunk,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    ChatMessage,
    FinetuningAssistantMessage,
    FinetuningMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.normalize import (
    InstructRequestNormalizer,
    InstructRequestNormalizerV7,
    InstructRequestNormalizerV13,
    InstructRequestNormalizerV15,
    get_normalizer,
)
from mistral_common.protocol.instruct.request import (
    ChatCompletionRequest,
    InstructRequest,
    ModelSettings,
    ReasoningEffort,
)
from mistral_common.protocol.instruct.tool_calls import Function, FunctionCall, Tool, ToolCall
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from mistral_common.tokens.tokenizers.model_settings_builder import EnumBuilder, ModelSettingsBuilder


def mock_chat_completion(messages: list[ChatMessage]) -> ChatCompletionRequest:
    return ChatCompletionRequest(
        model="test",
        messages=messages,
        top_p=1.0,
        temperature=0.7,
    )


class TestChatCompletionRequestNormalization:
    @pytest.fixture(autouse=True)
    def normalizer(self) -> InstructRequestNormalizer:
        return InstructRequestNormalizer(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        )

    def test_user_system_user(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content="U"),
                SystemMessage(content="S"),
                UserMessage(content="U"),
            ]
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="U"), UserMessage(content="U")],
            system_prompt="S",
        )

    def test_multiple_system(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                SystemMessage(content="S"),
                SystemMessage(content="S"),
                SystemMessage(content="S"),
            ]
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="")],
            system_prompt="S\n\nS\n\nS",
        )

    def test_single_system(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                SystemMessage(content="S"),
            ]
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="")],
            system_prompt="S",
        )

    def test_system_assistant_user(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                SystemMessage(content="S"),
                AssistantMessage(content="A"),
                UserMessage(content="U"),
            ]
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content=""), AssistantMessage(content="A"), UserMessage(content="U")],
            system_prompt="S",
        )

    def test_assistant_content_with_tool_calls(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(
                    content="A",
                    tool_calls=[ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "11"}'))],
                )
            ]
        )
        with pytest.raises(ValueError, match="Tool calls and content cannot be used together in the same message"):
            normalizer.from_chat_completion_request(chat_completion_request)

    def test_assistant_system_user_adds_user(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(content="A"),
                SystemMessage(content="S"),
                UserMessage(content="U"),
            ]
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content=""), AssistantMessage(content="A"), UserMessage(content="U")],
            system_prompt="S",
        )

    def test_message_aggregation_system_then_user(self, normalizer: InstructRequestNormalizer) -> None:
        parsed = normalizer.from_chat_completion_request(
            mock_chat_completion(
                [
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    UserMessage(content="u"),
                ]
            )
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="u")], system_prompt="s\n\ns\n\ns"
        )

    def test_message_aggregation_system_then_users(self, normalizer: InstructRequestNormalizer) -> None:
        parsed = normalizer.from_chat_completion_request(
            mock_chat_completion(
                [
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    UserMessage(content="u"),
                    UserMessage(content="u"),
                ]
            )
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="u\n\nu")], system_prompt="s\n\ns\n\ns"
        )

    def test_message_aggregation_mixed_with_middle_system(self, normalizer: InstructRequestNormalizer) -> None:
        parsed = normalizer.from_chat_completion_request(
            mock_chat_completion(
                [
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    UserMessage(content="u"),
                    UserMessage(content="u"),
                    SystemMessage(content="s"),
                    AssistantMessage(content="a"),
                    UserMessage(content="u"),
                ]
            )
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="u\n\nu"), AssistantMessage(content="a"), UserMessage(content="u")],
            system_prompt="s\n\ns\n\ns\n\ns",
        )

    def test_message_aggregation_consecutive_assistants(self, normalizer: InstructRequestNormalizer) -> None:
        parsed = normalizer.from_chat_completion_request(
            mock_chat_completion(
                [
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    SystemMessage(content="s"),
                    UserMessage(content="u"),
                    UserMessage(content="u"),
                    AssistantMessage(content="a"),
                    AssistantMessage(content="a"),
                    UserMessage(content="u"),
                ]
            )
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="u\n\nu"), AssistantMessage(content="a\n\na"), UserMessage(content="u")],
            system_prompt="s\n\ns\n\ns",
        )

    def test_message_aggregation_system_assistant_user(self, normalizer: InstructRequestNormalizer) -> None:
        parsed = normalizer.from_chat_completion_request(
            mock_chat_completion([SystemMessage(content="s"), AssistantMessage(content="a"), UserMessage(content="u")])
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content=""), AssistantMessage(content="a"), UserMessage(content="u")],
            system_prompt="s",
        )

    def test_tool_chunk_aggregation(self, normalizer: InstructRequestNormalizer) -> None:
        messages = [
            ToolMessage(content="C", tool_call_id="1"),
            ToolMessage(content=[TextChunk(text='{"a": 2}')], tool_call_id="2"),
            ToolMessage(content=[TextChunk(text="B"), TextChunk(text="A")], tool_call_id="3"),
        ]

        expected = [
            ToolMessage(content="C", tool_call_id="1"),
            ToolMessage(content=json.dumps({"a": 2}), tool_call_id="2"),
            ToolMessage(content="B\n\nA", tool_call_id="3"),
        ]

        assert (
            normalizer._aggregate_tool_messages(
                messages, [tool_message.tool_call_id for tool_message in messages if tool_message.tool_call_id]
            )
            == expected
        )

    def test_normalize_chunks(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content="foo"),
                UserMessage(
                    content=[TextChunk(type=ChunkTypes.text, text="chunk")],
                ),
                UserMessage(content="foo"),
                UserMessage(
                    content=[TextChunk(type=ChunkTypes.text, text="chunk")],
                ),
            ],
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="foo\n\nchunk\n\nfoo\n\nchunk")]
        )

    def test_many_chunks_in_user_message(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content="foo"),
                UserMessage(
                    content=[
                        TextChunk(type=ChunkTypes.text, text="chunk1"),
                        TextChunk(type=ChunkTypes.text, text="chunk2"),
                        TextChunk(type=ChunkTypes.text, text="chunk3"),
                    ],
                ),
            ],
        )
        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="foo\n\nchunk1\n\nchunk2\n\nchunk3")]
        )

    def test_ignore_middle_empty_text_chunks(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(
                    content=[
                        TextChunk(text="U"),
                        TextChunk(text=""),
                        TextChunk(text="V"),
                    ],
                ),
                AssistantMessage(
                    content=[
                        TextChunk(text="A"),
                        TextChunk(text=""),
                        TextChunk(text="B"),
                    ],
                ),
            ]
        )
        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="U\n\nV"), AssistantMessage(content="A\n\nB")],
        )

    def test_safety_prompt_aggregated(self, normalizer: InstructRequestNormalizer) -> None:
        chat_completion_request = ChatCompletionRequest[ChatMessage](
            model="test",
            messages=[
                UserMessage(content="user"),
                SystemMessage(content="system"),
            ],
            top_p=1.0,
            temperature=0.7,
        )

        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="user")],
            system_prompt="system",
        )

    def test_normalize_tools(self, normalizer: InstructRequestNormalizer) -> None:
        """
        Test doesn't really "normalize" anything but it checks that the tools are added to the
        InstructRequest during from_chat_completion_request
        """
        tools = [
            Tool(function=Function(name="tool1", description="1", parameters={})),
            Tool(function=Function(name="tool2", description="2", parameters={})),
        ]

        request = ChatCompletionRequest[ChatMessage](
            model="triton",
            messages=[
                SystemMessage(content="helpful assistant"),
                UserMessage(content="a"),
                AssistantMessage(content="b"),
                UserMessage(content="c"),
            ],
            tools=tools,
        )

        gt = InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="a"), AssistantMessage(content="b"), UserMessage(content="c")],
            available_tools=tools,
            system_prompt="helpful assistant",
        )

        normalized = normalizer.from_chat_completion_request(request)
        assert normalized == gt

    def test_normalize_funcalls(self, normalizer: InstructRequestNormalizer) -> None:
        request = ChatCompletionRequest[ChatMessage](
            model="triton",
            messages=[
                SystemMessage(content="helpful assistant"),
                UserMessage(content="a"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(
                            function=FunctionCall(name="tool1", arguments='\n\n{\n"input":\n      "11"\n}\n\n')
                        ),  # dodgy input json string
                    ],
                ),
                ToolMessage(
                    name="tool1",
                    content='{\n"output":\n      "11"\n}',  # dodgy input json string
                ),
            ],
            tools=[Tool(function=Function(name="tool1", description="1", parameters={}))],
        )

        gt = InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="a"),
                AssistantMessage(
                    content="",
                    tool_calls=[
                        ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "11"}')),  # clean json string
                    ],
                ),
                ToolMessage(
                    name="tool1",
                    content='{"output": "11"}',  # clean json string
                ),
            ],
            system_prompt="helpful assistant",
            available_tools=[Tool(function=Function(name="tool1", description="1", parameters={}))],
        )

        normalized = normalizer.from_chat_completion_request(request)
        assert normalized == gt

    def test_assert_parsed_settings(
        self,
        normalizer: InstructRequestNormalizer,
    ) -> None:
        chat_completion_request = ChatCompletionRequest(messages=[UserMessage(content="B")])
        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
        assert parsed_request == InstructRequest[ChatMessage, Tool](messages=[UserMessage(content="B")])

    def test_continue_final_message_forwarded(self, normalizer: InstructRequestNormalizer) -> None:
        request = ChatCompletionRequest[ChatMessage](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )
        result = normalizer.from_chat_completion_request(request)
        assert result == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )

    def test_json_normalizes_tool_content(self, normalizer: InstructRequestNormalizer) -> None:
        r"""Base normalizer (v1-v3) JSON-normalizes tool message content."""
        messy_json = '{"key" :  "value" ,  "num": 1}'
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")]),
                ToolMessage(content=messy_json, tool_call_id="c1"),
            ],
        )
        parsed = normalizer.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")],
                ),
                ToolMessage(content='{"key": "value", "num": 1}', tool_call_id="c1"),
            ],
        )


class TestChatCompletionRequestNormalizationV7:
    @pytest.fixture(autouse=True)
    def normalizer_v7(self) -> InstructRequestNormalizerV7:
        return InstructRequestNormalizerV7(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        )

    def test_system_assistant_user_v7(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                SystemMessage(content="S"),
                AssistantMessage(content="A"),
                UserMessage(content="U"),
            ]
        )

        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[SystemMessage(content="S"), AssistantMessage(content="A"), UserMessage(content="U")],
        )

    def test_assistant_assistant_system_v7(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(content="A"),
                SystemMessage(content="S"),
            ]
        )

        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[AssistantMessage(content="A"), SystemMessage(content="S")],
        )

    def test_assistant_content_with_tool_calls(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(
                    content="A",
                    tool_calls=[ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "11"}'))],
                )
            ]
        )
        normalized_chat_req: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert normalized_chat_req == InstructRequest[ChatMessage, Tool](
            messages=[
                AssistantMessage(
                    content="A",
                    tool_calls=[ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "11"}'))],
                ),
            ],
        )

    def test_assistant_content_with_more_tool_calls(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content="A1"),
                AssistantMessage(content="B1"),
                AssistantMessage(
                    content="B2",
                    tool_calls=[ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "1"}'))],
                ),
                AssistantMessage(content="B3"),
                AssistantMessage(
                    content="B4",
                    tool_calls=[
                        ToolCall(function=FunctionCall(name="tool21", arguments='{"input": "21"}')),
                        ToolCall(function=FunctionCall(name="tool22", arguments='{"input": "22"}')),
                    ],
                ),
                AssistantMessage(content="B5"),
                UserMessage(content="C1"),
            ]
        )
        normalized_chat_req: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert normalized_chat_req == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A1"),
                AssistantMessage(
                    content="B1\n\nB2\n\nB3\n\nB4\n\nB5",
                    tool_calls=[
                        ToolCall(function=FunctionCall(name="tool1", arguments='{"input": "1"}')),
                        ToolCall(function=FunctionCall(name="tool21", arguments='{"input": "21"}')),
                        ToolCall(function=FunctionCall(name="tool22", arguments='{"input": "22"}')),
                    ],
                ),
                UserMessage(content="C1"),
            ],
        )

    def test_assert_parsed_settings(
        self,
        normalizer_v7: InstructRequestNormalizerV7,
    ) -> None:
        chat_completion_request = ChatCompletionRequest(messages=[UserMessage(content="B")])
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](messages=[UserMessage(content="B")])

    def test_continue_final_message_forwarded(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        request = ChatCompletionRequest[ChatMessage](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )
        result: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(request)
        assert result == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )

    @pytest.mark.parametrize("num_empty", [0, 1, 2])
    def test_only_empty_text_chunks(self, normalizer_v7: InstructRequestNormalizerV7, num_empty: int) -> None:
        """Messages with only empty TextChunks produce empty content."""
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content=[TextChunk(text="") for _ in range(num_empty)]),
                AssistantMessage(content=[TextChunk(text="") for _ in range(num_empty)]),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content=""), AssistantMessage(content="")],
        )

    def test_complex_user_aggregation(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        """Complex multi-user-message aggregation with mixed str, chunks, empty, and non-text chunks."""

        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content=""),
                UserMessage(content="A"),
                UserMessage(content=""),
                UserMessage(content=[TextChunk(text="B")]),
                UserMessage(content=[TextChunk(text="")]),
                UserMessage(
                    content=[
                        TextChunk(text="C"),
                        TextChunk(text="D"),
                        ImageURLChunk(image_url="E"),
                        TextChunk(text="G"),
                        TextChunk(text=""),
                        TextChunk(text="H"),
                    ]
                ),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(
                    content=[
                        TextChunk(text="A\n\nB\n\nC\n\nD"),
                        ImageURLChunk(image_url="E"),
                        TextChunk(text="G\n\nH"),
                    ]
                ),
            ],
        )

    def test_complex_assistant_aggregation(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        """Complex multi-assistant-message aggregation with mixed str, chunks, and empty content."""
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(content=""),
                AssistantMessage(content="A"),
                AssistantMessage(content=""),
                AssistantMessage(content=[TextChunk(text="B")]),
                AssistantMessage(content=[TextChunk(text="")]),
                AssistantMessage(
                    content=[
                        TextChunk(text="C"),
                        TextChunk(text="D"),
                    ]
                ),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[AssistantMessage(content="A\n\nB\n\nC\n\nD")],
        )

    def test_accepts_string_content(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        r"""V7 normalizer accepts string content in assistant messages."""
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content="plain text"),
            ],
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="query"), AssistantMessage(content="plain text")],
        )

    def test_skips_json_normalization_on_tool_content(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        r"""V7+ normalizers do not JSON-normalize tool message content."""
        messy_json = '{"key" :  "value" ,  "num": 1}'
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")]),
                ToolMessage(content=messy_json, tool_call_id="c1"),
            ],
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")],
                ),
                ToolMessage(content=messy_json, tool_call_id="c1"),
            ],
        )

    def test_preserves_audio_in_system_message(self, normalizer_v7: InstructRequestNormalizerV7) -> None:
        r"""V7 normalizer preserves AudioChunk in system messages."""
        request = mock_chat_completion(
            messages=[
                SystemMessage(content=[TextChunk(text="hello"), AudioChunk(input_audio=b"fake_audio_data")]),
                UserMessage(content="test"),
            ]
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v7.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                SystemMessage(
                    content=[TextChunk(text="hello"), AudioChunk(input_audio=b"fake_audio_data")],
                ),
                UserMessage(content="test"),
            ],
        )


class TestFineTuningNormalizer:
    @pytest.fixture(autouse=True)
    def normalizer(self) -> InstructRequestNormalizer:
        return InstructRequestNormalizer(
            UserMessage, FinetuningAssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        )

    def test_normalize_weighted_assistant(self, normalizer: InstructRequestNormalizer) -> None:
        request = ChatCompletionRequest[FinetuningMessage](
            messages=[
                SystemMessage(content="helpful assistant"),
                UserMessage(content="a"),
                FinetuningAssistantMessage(content="he", weight=0),
                UserMessage(content="b"),
                FinetuningAssistantMessage(content="ho", weight=1),
                FinetuningAssistantMessage(content="lla", weight=1),
            ],
        )
        expected = InstructRequest[FinetuningMessage, Tool](
            messages=[
                UserMessage(content="a"),
                FinetuningAssistantMessage(content="he", weight=0),
                UserMessage(content="b"),
                FinetuningAssistantMessage(content="ho\n\nlla", weight=1),
            ],
            system_prompt="helpful assistant",
        )
        normalized = normalizer.from_chat_completion_request(request)
        assert expected == normalized

    def test_should_not_aggregate_if_weight_is_different(self, normalizer: InstructRequestNormalizer) -> None:
        request = ChatCompletionRequest[FinetuningMessage](
            messages=[
                SystemMessage(content="helpful assistant"),
                UserMessage(content="a"),
                FinetuningAssistantMessage(content="he", weight=0),
                UserMessage(content="b"),
                FinetuningAssistantMessage(content="ho", weight=0),
                FinetuningAssistantMessage(content="lla", weight=1),
            ],
        )
        expected = InstructRequest[FinetuningMessage, Tool](
            messages=[
                UserMessage(content="a"),
                FinetuningAssistantMessage(content="he", weight=0),
                UserMessage(content="b"),
                FinetuningAssistantMessage(content="ho", weight=0),
                FinetuningAssistantMessage(content="lla", weight=1),
            ],
            system_prompt="helpful assistant",
        )
        normalized = normalizer.from_chat_completion_request(request)
        assert normalized == expected


class TestChatCompletionRequestNormalizationV13:
    @pytest.fixture(autouse=True)
    def normalizer_v13(self) -> InstructRequestNormalizerV13:
        return InstructRequestNormalizerV13(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest, None
        )

    def _mock_chat_completion(self, messages: list[ChatMessage]) -> ChatCompletionRequest:
        return ChatCompletionRequest(
            model="test",
            messages=messages,
            top_p=1.0,
            temperature=0.7,
        )

    def test_no_reorder_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
            ],
        )

    def test_reorder_last_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="D", tool_call_id="2"),
                ToolMessage(content="C", tool_call_id="1"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
            ],
        )

    def test_reorder_internal_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="D", tool_call_id="2"),
                ToolMessage(content="C", tool_call_id="1"),
                AssistantMessage(content="E"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
                AssistantMessage(content="E"),
            ],
        )

    def test_reorder_extra_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="D", tool_call_id="2"),
                ToolMessage(content="C", tool_call_id="1"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
            ],
        )

    def test_reorder_only_from_latest_assistant_message(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
                AssistantMessage(
                    content="E",
                    tool_calls=[
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="A"),
                AssistantMessage(
                    content="B",
                    tool_calls=[
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="C", tool_call_id="1"),
                ToolMessage(content="D", tool_call_id="2"),
                AssistantMessage(
                    content="E",
                    tool_calls=[
                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
                    ],
                ),
                ToolMessage(content="D", tool_call_id="2"),
                ToolMessage(content="C", tool_call_id="1"),
            ],
        )

    @pytest.mark.parametrize(
        ["system_message", "expected_system_message"],
        [
            (
                SystemMessage(content="A"),
                SystemMessage(content="A"),
            ),
            (
                SystemMessage(content=[TextChunk(text="A")]),
                SystemMessage(content="A"),
            ),
            (
                SystemMessage(content=[TextChunk(text="A"), TextChunk(text="B")]),
                SystemMessage(content="A\n\nB"),
            ),
            (
                SystemMessage(content=[TextChunk(text="A"), TextChunk(text="B"), ThinkChunk(thinking="C")]),
                SystemMessage(content=[TextChunk(text="A\n\nB"), ThinkChunk(thinking="C")]),
            ),
            (
                SystemMessage(
                    content=[
                        TextChunk(text="A"),
                        TextChunk(text="B"),
                        ThinkChunk(thinking="C"),
                        ThinkChunk(thinking="D"),
                    ]
                ),
                SystemMessage(content=[TextChunk(text="A\n\nB"), ThinkChunk(thinking="C"), ThinkChunk(thinking="D")]),
            ),
        ],
    )
    def test_aggregate_system_prompt_content(
        self,
        normalizer_v13: InstructRequestNormalizerV13,
        system_message: SystemMessage,
        expected_system_message: SystemMessage,
    ) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
            messages=[system_message, UserMessage(content="B")]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[expected_system_message, UserMessage(content="B")]
        )

    def test_system_messages_no_aggregation(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        """Consecutive system messages are NOT aggregated into one in V7+."""
        chat_completion_request = self._mock_chat_completion(
            messages=[
                SystemMessage(content="A"),
                SystemMessage(content="B"),
                UserMessage(content="C"),
                SystemMessage(content="D"),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                SystemMessage(content="A"),
                SystemMessage(content="B"),
                UserMessage(content="C"),
                SystemMessage(content="D"),
            ],
        )

    def test_system_messages_normalization(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        """System message chunks within the same message are aggregated with no separator."""
        chat_completion_request = self._mock_chat_completion(
            messages=[
                SystemMessage(content=[TextChunk(text="A"), TextChunk(text="B")]),
                SystemMessage(content=[TextChunk(text="C")]),
            ]
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[
                SystemMessage(content="A\n\nB"),
                SystemMessage(content="C"),
            ],
        )

    def test_assert_parsed_settings(
        self,
        normalizer_v13: InstructRequestNormalizerV13,
    ) -> None:
        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(messages=[UserMessage(content="B")])
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](messages=[UserMessage(content="B")])

    def test_continue_final_message_forwarded(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        request = ChatCompletionRequest[ChatMessage](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )
        result: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(request)
        assert result == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
        )

    def test_accepts_text_and_think_chunks(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        r"""V13 normalizer accepts TextChunk and ThinkChunk in assistant messages."""
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content=[ThinkChunk(thinking="reasoning"), TextChunk(text="answer")]),
            ],
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content=[ThinkChunk(thinking="reasoning"), TextChunk(text="answer")]),
            ],
        )

    def test_accepts_string_content(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        r"""V13 normalizer accepts string content in assistant messages."""
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content="plain text"),
            ],
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="query"), AssistantMessage(content="plain text")],
        )

    def test_assistant_think_chunk_inter_message_aggregation(
        self, normalizer_v13: InstructRequestNormalizerV13
    ) -> None:
        r"""V13 normalizer preserves ThinkChunks across multiple assistant messages."""
        chat_completion_request = mock_chat_completion(
            messages=[
                AssistantMessage(content="A"),
                AssistantMessage(content=[TextChunk(text="B")]),
                AssistantMessage(
                    content=[
                        ThinkChunk(thinking="T"),
                        TextChunk(text="C"),
                        TextChunk(text="D"),
                    ]
                ),
            ]
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                AssistantMessage(
                    content=[
                        TextChunk(text="A\n\nB"),
                        ThinkChunk(thinking="T"),
                        TextChunk(text="C\n\nD"),
                    ]
                ),
            ],
        )

    def test_assistant_think_chunk_intra_message_aggregation(
        self, normalizer_v13: InstructRequestNormalizerV13
    ) -> None:
        r"""V13 normalizer coalesces TextChunks and preserves multiple ThinkChunks within a single message."""
        chat_completion_request = mock_chat_completion(
            messages=[
                UserMessage(content="u"),
                AssistantMessage(
                    content=[
                        ThinkChunk(thinking="t1"),
                        ThinkChunk(thinking="t2"),
                        TextChunk(text="a1"),
                        TextChunk(text="a2"),
                        TextChunk(text="a3"),
                    ]
                ),
                UserMessage(content="u"),
            ]
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="u"),
                AssistantMessage(
                    content=[
                        ThinkChunk(thinking="t1"),
                        ThinkChunk(thinking="t2"),
                        TextChunk(text="a1\n\na2\n\na3"),
                    ]
                ),
                UserMessage(content="u"),
            ],
        )

    def test_aggregates_text_tool_content(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
        r"""V13 normalizer aggregates TextChunks in tool messages to a string."""
        request = mock_chat_completion(
            messages=[
                UserMessage(content="query"),
                AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")]),
                ToolMessage(content=[TextChunk(text="hello"), TextChunk(text="world")], tool_call_id="c1"),
            ],
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")],
                ),
                ToolMessage(content="hello\n\nworld", tool_call_id="c1"),
            ],
        )


class TestChatCompletionRequestNormalizationV15:
    @pytest.fixture(autouse=True)
    def normalizer_v15(self) -> InstructRequestNormalizerV15:
        return InstructRequestNormalizerV15(
            UserMessage,
            AssistantMessage,
            ToolMessage,
            SystemMessage,
            InstructRequest,
            ModelSettingsBuilder(
                reasoning_effort=EnumBuilder[ReasoningEffort](
                    values=list(ReasoningEffort), accepts_none=False, default=None
                )
            ),
        )

    @pytest.mark.parametrize("reasoning_effort", [ReasoningEffort.none, ReasoningEffort.high])
    def test_assert_parsed_settings(
        self, normalizer_v15: InstructRequestNormalizerV15, reasoning_effort: ReasoningEffort
    ) -> None:
        chat_completion_request = ChatCompletionRequest(
            messages=[UserMessage(content="B")], reasoning_effort=reasoning_effort
        )
        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(
            chat_completion_request
        )
        assert parsed_request == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="B")],
            settings=ModelSettings(reasoning_effort=reasoning_effort),
        )

    def test_continue_final_message_forwarded(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        request = ChatCompletionRequest[ChatMessage](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
            reasoning_effort=ReasoningEffort.high,
        )
        result: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert result == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="a"), AssistantMessage(content="b")],
            continue_final_message=True,
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_v15_intra_message_chunks_joined_without_separator(
        self, normalizer_v15: InstructRequestNormalizerV15
    ) -> None:
        """V15 joins TextChunks within the same message with no separator (chunk_join_str='')."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content=[TextChunk(text="A"), TextChunk(text="B")]),
                AssistantMessage(content=[TextChunk(text="C"), TextChunk(text="D")]),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="AB"), AssistantMessage(content="CD")],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_v15_inter_message_join_still_uses_separator(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 still joins text across different messages with '\n\n'."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content="First"),
                UserMessage(content="Second"),
                AssistantMessage(content="Reply"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="First\n\nSecond"), AssistantMessage(content="Reply")],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_v15_mixed_intra_and_inter_message(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 combines intra-message ('') and inter-message ('\n\n') joining."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content=[TextChunk(text="A"), TextChunk(text="B")]),
                UserMessage(content=[TextChunk(text="C"), TextChunk(text="D")]),
                AssistantMessage(content="Reply"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="AB\n\nCD"), AssistantMessage(content="Reply")],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_v15_mixed_intra_and_inter_assistant_messages(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 combines intra-message ('') and inter-message ('\n\n') joining for assistant messages."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content="Hello"),
                AssistantMessage(content=[TextChunk(text="A"), TextChunk(text="B")]),
                AssistantMessage(content=[TextChunk(text="C"), TextChunk(text="D")]),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="Hello"), AssistantMessage(content="AB\n\nCD")],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_v15_tool_message_text_chunks_joined_without_separator(
        self, normalizer_v15: InstructRequestNormalizerV15
    ) -> None:
        """V15 tool message TextChunks are joined with no separator."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")]),
                ToolMessage(content=[TextChunk(text="X"), TextChunk(text="Y")], tool_call_id="c1"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")],
                ),
                ToolMessage(content="XY", tool_call_id="c1"),
            ],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_accepts_text_and_think_chunks(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 normalizer accepts TextChunk and ThinkChunk in assistant messages."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content=[ThinkChunk(thinking="reasoning"), TextChunk(text="answer")]),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content=[ThinkChunk(thinking="reasoning"), TextChunk(text="answer")]),
            ],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_accepts_string_content(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 normalizer accepts string content in assistant messages."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(content="plain text"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[UserMessage(content="query"), AssistantMessage(content="plain text")],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_preserves_non_text_tool_content(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 normalizer preserves non-text chunks in tool messages."""
        image_chunk = ImageURLChunk(image_url="https://example.com/image.png")
        request = ChatCompletionRequest(  # type: ignore[type-var]
            messages=[
                UserMessage(content="query"),
                AssistantMessage(tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")]),
                ToolMessage(content=[image_chunk], tool_call_id="c1"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[ToolCall(function=FunctionCall(name="fn", arguments="{}"), id="c1")],
                ),
                ToolMessage(content=[image_chunk], tool_call_id="c1"),
            ],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_sorts_multimodal_tool_messages(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 normalizer sorts multimodal tool messages by tool call order."""
        image_chunk_1 = ImageURLChunk(image_url="https://example.com/img1.png")
        image_chunk_2 = ImageURLChunk(image_url="https://example.com/img2.png")
        request = ChatCompletionRequest(  # type: ignore[type-var]
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    tool_calls=[
                        ToolCall(function=FunctionCall(name="fn1", arguments="{}"), id="c1"),
                        ToolCall(function=FunctionCall(name="fn2", arguments="{}"), id="c2"),
                    ]
                ),
                ToolMessage(content=[image_chunk_2], tool_call_id="c2"),
                ToolMessage(content=[image_chunk_1], tool_call_id="c1"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                UserMessage(content="query"),
                AssistantMessage(
                    content="",
                    tool_calls=[
                        ToolCall(function=FunctionCall(name="fn1", arguments="{}"), id="c1"),
                        ToolCall(function=FunctionCall(name="fn2", arguments="{}"), id="c2"),
                    ],
                ),
                ToolMessage(content=[image_chunk_1], tool_call_id="c1"),
                ToolMessage(content=[image_chunk_2], tool_call_id="c2"),
            ],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )

    def test_preserves_audio_in_system_message(self, normalizer_v15: InstructRequestNormalizerV15) -> None:
        r"""V15 normalizer preserves AudioChunk in system messages."""
        request = ChatCompletionRequest[ChatMessage](
            messages=[
                SystemMessage(content=[TextChunk(text="hello"), AudioChunk(input_audio=b"fake_audio_data")]),
                UserMessage(content="test"),
            ],
            reasoning_effort=ReasoningEffort.high,
        )
        parsed: InstructRequest[ChatMessage, Tool] = normalizer_v15.from_chat_completion_request(request)
        assert parsed == InstructRequest[ChatMessage, Tool](
            messages=[
                SystemMessage(
                    content=[TextChunk(text="hello"), AudioChunk(input_audio=b"fake_audio_data")],
                ),
                UserMessage(content="test"),
            ],
            settings=ModelSettings(reasoning_effort=ReasoningEffort.high),
        )


@pytest.mark.parametrize(
    "version,expected_class,model_settings_builder",
    [
        (TokenizerVersion.v1, InstructRequestNormalizer, None),
        (TokenizerVersion.v2, InstructRequestNormalizer, None),
        (TokenizerVersion.v3, InstructRequestNormalizer, None),
        (TokenizerVersion.v7, InstructRequestNormalizerV7, None),
        (TokenizerVersion.v11, InstructRequestNormalizerV7, None),
        (TokenizerVersion.v13, InstructRequestNormalizerV13, None),
        (
            TokenizerVersion.v15,
            InstructRequestNormalizerV15,
            ModelSettingsBuilder(
                reasoning_effort=EnumBuilder[ReasoningEffort](
                    values=list(ReasoningEffort), accepts_none=False, default=None
                )
            ),
        ),
    ],
)
def test_get_normalizer_version_mapping(
    version: TokenizerVersion, expected_class: type, model_settings_builder: ModelSettingsBuilder
) -> None:
    normalizer = get_normalizer(version, model_settings_builder)
    assert isinstance(normalizer, expected_class)
    assert normalizer._model_settings_builder == model_settings_builder



import pytest
from pydantic import ValidationError

from mistral_common.base import MistralBase
from mistral_common.protocol.instruct.chunk import TextChunk
from mistral_common.protocol.instruct.messages import UserMessage


def test_filter_cls_fields() -> None:
    assert MistralBase._filter_cls_fields({}) == {}

    filtered = UserMessage._filter_cls_fields({"role": "user", "content": "hi", "name": "u1"})
    assert filtered == {"role": "user", "content": "hi"}

    filtered = TextChunk._filter_cls_fields({"type": "text", "text": "hi", "annotations": []})
    assert filtered == {"type": "text", "text": "hi"}


def test_model_validate_ignore_extra_filters_and_validates() -> None:
    message = UserMessage.model_validate_ignore_extra({"role": "user", "content": "hi", "name": "u1"})
    assert message == UserMessage(content="hi")


def test_model_validate_ignore_extra_no_extra_keys() -> None:
    chunk = TextChunk.model_validate_ignore_extra({"type": "text", "text": "hello"})
    assert chunk == TextChunk(text="hello")


def test_model_validate_ignore_extra_raises_on_missing_required() -> None:
    with pytest.raises(ValidationError):
        UserMessage.model_validate_ignore_extra({"role": "user"})



import json
import warnings
from typing import Generic, Sequence

from typing_extensions import assert_never

from mistral_common.exceptions import InvalidRequestException
from mistral_common.protocol.instruct.chunk import (
    ContentChunk,
    TextChunk,
)
from mistral_common.protocol.instruct.messages import (
    UATS,
    AssistantMessage,
    AssistantMessageType,
    FinetuningAssistantMessage,
    Roles,
    SystemMessage,
    SystemMessageType,
    ToolMessage,
    ToolMessageType,
    UserMessage,
    UserMessageType,
)
from mistral_common.protocol.instruct.request import ChatCompletionRequest, InstructRequest, ModelSettings
from mistral_common.protocol.instruct.tool_calls import FunctionCall, Tool, ToolCall
from mistral_common.tokens.tokenizers.base import InstructRequestType, TokenizerVersion
from mistral_common.tokens.tokenizers.model_settings_builder import ModelSettingsBuilder

_DEFAULT_JOIN_STR = "\n\n"


def _aggregate_content_chunks_impl(
    contents: list[list[ContentChunk] | str | None],
    msg_join_str: str,
    chunk_join_str: str,
) -> list[ContentChunk] | str:
    r"""Coalesce TextChunks within the same message and across different messages.

    Adjacent TextChunks within the same message are joined with `chunk_join_str`.
    Text from different messages is joined with `msg_join_str`.

    Args:
        contents: A list of message contents, where each element is either a string,
            a list of ContentChunks, or None. This is typically
            `[message.content for message in messages]`.
        msg_join_str: Separator inserted between text from different messages.
        chunk_join_str: Separator inserted between adjacent text chunks within
            the same message.

    Returns:
        A plain string if only text chunks were present, otherwise a list of
        ContentChunks with adjacent text coalesced.
    """
    all_content: list[ContentChunk] = []
    cur_text_parts: list[str] = []

    def _flush_text() -> None:
        if cur_text_parts:
            all_content.append(TextChunk(text="".join(cur_text_parts)))
            cur_text_parts.clear()

    for content in contents:
        needs_new_msg_sep = bool(cur_text_parts)
        if not content:  # skip None or empty string
            continue
        elif isinstance(content, str):
            join = msg_join_str if needs_new_msg_sep else ""
            cur_text_parts.append(join + content)
        else:  # list[ContentChunk]
            for chunk in content:
                if isinstance(chunk, TextChunk):
                    if not chunk.text:  # skip empty text chunks
                        continue
                    if not cur_text_parts:
                        join = ""
                    elif needs_new_msg_sep:
                        join = msg_join_str
                    else:
                        join = chunk_join_str
                    cur_text_parts.append(join + chunk.text)
                    needs_new_msg_sep = False
                else:
                    _flush_text()
                    all_content.append(chunk)

    if not all_content:
        # Only text encountered: return as str
        return "".join(cur_text_parts)

    _flush_text()

    return all_content


class InstructRequestNormalizer(
    Generic[UserMessageType, AssistantMessageType, ToolMessageType, SystemMessageType, InstructRequestType]
):
    r"""Takes a [ChatCompletionRequest][mistral_common.protocol.instruct.request.ChatCompletionRequest] and normalizes
    it into an [InstructRequest][mistral_common.protocol.instruct.request.InstructRequest].

    The normalization process does several things such as:
    - Aggregate consecutive messages of the same role
    - Aggregate system prompts
    - Normalize json content
    - Normalize tool calls

    Examples:
        >>> normalizer = InstructRequestNormalizer.normalizer()
    """

    _system_prompt_in_begin: bool = False
    _allow_tool_call_and_content: bool = False
    _chunk_join_str: str = _DEFAULT_JOIN_STR
    _msg_join_str: str = _DEFAULT_JOIN_STR

    def __init__(
        self,
        user_message_class: type[UserMessageType],
        assistant_message_class: type[AssistantMessageType],
        tool_message_class: type[ToolMessageType],
        system_message_class: type[SystemMessageType],
        instruct_request_class: type[InstructRequestType],
        model_settings_builder: ModelSettingsBuilder | None,
    ):
        r"""Initializes the normalizer with the appropriate message classes.

        Args:
           user_message_class: The class for user messages.
           assistant_message_class: The class for assistant messages.
           tool_message_class: The class for tool messages.
           system_message_class: The class for system messages.
           instruct_request_class: The class for instruct requests.
           model_settings_builder: The builder for model settings, or None if unsupported.
        """
        self._user_message_class = user_message_class
        self._assistant_message_class = assistant_message_class
        self._tool_message_class = tool_message_class
        self._instruct_request_class = instruct_request_class
        # this is unused but makes creation nicer
        self._system_message_class = system_message_class
        self._model_settings_builder = model_settings_builder

    @staticmethod
    def normalizer(model_settings_builder: ModelSettingsBuilder | None = None) -> "InstructRequestNormalizer":
        r"""Returns a normalizer for the default instruct request.

        Args:
            model_settings_builder: Must be None for this normalizer version.

        Returns:
            A normalizer for the default instruct request.

        Raises:
            ValueError: If model_settings_builder is not None.

        Examples:
            >>> normalizer = InstructRequestNormalizer.normalizer()
        """
        if model_settings_builder is not None:
            raise ValueError(
                f"model_settings_builder must be None for InstructRequestNormalizer, got {model_settings_builder}"
            )
        return InstructRequestNormalizer(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest[UATS, Tool], None
        )

    def build_settings(self, request: ChatCompletionRequest) -> ModelSettings:
        r"""Build model settings from a chat completion request.

        For pre-v15 normalizers, model settings are all `None`.

        Args:
            request: The chat completion request.

        Returns:
            Returns `ModelSettings.none()`.
        """
        if self._model_settings_builder is not None:
            raise InvalidRequestException(
                f"model_settings_builder should be None for {type(self).__name__}, got {self._model_settings_builder}"
            )
        return ModelSettings.none()

    def _normalize_json_content(self, content: str | None) -> str:
        if content is None or len(content) == 0:
            return "{}"

        try:
            parsed_json = json.loads(content)
            normalized_content = json.dumps(parsed_json, ensure_ascii=False)
        except json.JSONDecodeError:
            normalized_content = content
        return normalized_content

    def _aggregate_content_chunks(self, messages: list[UATS]) -> list[ContentChunk] | str:
        """Coalesce neighboring blocks of ContentChunks across messages."""
        return _aggregate_content_chunks_impl(
            [message.content for message in messages],
            msg_join_str=self._msg_join_str,
            chunk_join_str=self._chunk_join_str,
        )

    def _aggregate_content_chunks_to_str_same_message(self, message: UATS) -> str:
        """Aggregate a single message's content chunks to a string.

        Args:
            message: A single message with role system or tool.

        Returns:
            The aggregated content as a string.
        """
        assert message.role in (Roles.system, Roles.tool), message.role
        aggregated = self._aggregate_content_chunks([message])
        assert isinstance(aggregated, str), aggregated
        return aggregated

    def _aggregate_system_prompts(self, messages: list[UATS]) -> str | None:
        system_prompt: list[str] = []

        for message in messages:
            if message.role == Roles.system and message.content:
                system_prompt.append(self._aggregate_content_chunks_to_str_same_message(message))

        return self._msg_join_str.join(system_prompt) if len(system_prompt) else None

    def _aggregate_tool_messages(self, messages: list[UATS], latest_call_ids: list[str]) -> list[ToolMessageType]:
        """Normalize tool messages without aggregation across messages.

        Each tool message's content is JSON-normalized; chunk types are guaranteed by the validator.
        """
        tool_messages: list[ToolMessageType] = []
        for message in messages:
            assert isinstance(message, self._tool_message_class), "Expected tool message"
            content = self._aggregate_content_chunks([message])
            assert isinstance(content, str), (
                f"Unexpected content chunk types in tool message: {[type(c).__name__ for c in content]}"
            )
            normalized_content = self._normalize_json_content(content)

            tool_messages.append(
                self._tool_message_class(
                    content=normalized_content, tool_call_id=message.tool_call_id, name=message.name
                )
            )

        return tool_messages

    def _normalize_tool_call(self, tool_call: ToolCall) -> ToolCall:
        normalized_function_aruments = self._normalize_json_content(tool_call.function.arguments)
        return ToolCall(
            function=FunctionCall(name=tool_call.function.name, arguments=normalized_function_aruments),
            id=tool_call.id,
        )

    def _aggregate_system_messages(self, messages: list[UATS]) -> list[SystemMessageType]:
        return []

    def _aggregate_assistant_messages(self, messages: list[UATS]) -> AssistantMessageType:
        tool_calls: list[ToolCall] = []
        prefix: bool = False
        weight: float | None = None

        content = self._aggregate_content_chunks(messages)

        for message in messages:
            assert isinstance(message, self._assistant_message_class), "Expected assistant message"

            if not self._allow_tool_call_and_content and (message.tool_calls and message.content):
                raise ValueError(f"Tool calls and content cannot be used together in the same message. {message}")

            if message.tool_calls:
                for tool_call in message.tool_calls:
                    normalized_tool_call = self._normalize_tool_call(tool_call)
                    tool_calls.append(normalized_tool_call)

            prefix |= message.prefix

            if isinstance(message, FinetuningAssistantMessage):
                if weight is not None:
                    assert weight == message.weight, (
                        "Expected weights of aggregated FinetuningAssistantMessage to be equal"
                    )
                weight = message.weight

        aggregated_message = self._assistant_message_class(
            content=content,
            tool_calls=tool_calls or None,
            prefix=prefix,
        )

        if weight is not None and hasattr(aggregated_message, "weight"):
            aggregated_message.weight = weight
        return aggregated_message

    def _aggregate_user_messages(self, messages: list[UATS]) -> UserMessageType:
        """Coalesce neighboring blocks of ContentChunks in user messages."""
        content = self._aggregate_content_chunks(messages)
        return self._user_message_class(content=content)

    def _aggregate_role(self, messages: list[UATS], role: Roles | None, latest_call_ids: list[str]) -> Sequence[UATS]:
        if role == Roles.tool:
            return self._aggregate_tool_messages(messages, latest_call_ids)
        elif role == Roles.assistant:
            return [self._aggregate_assistant_messages(messages)]
        elif role == Roles.user:
            return [self._aggregate_user_messages(messages)]
        else:  # System messages are ignored
            return self._aggregate_system_messages(messages)

    def _aggregate_messages(self, messages: list[UATS]) -> list[UATS]:
        aggregated_messages: list[UATS] = []
        messages_to_aggregate: list[UATS] = []
        current_role: Roles | None = None
        current_weight: float | None = None
        latest_call_ids: list[str] = []

        # Collect consecutive lists of messages with the same role and weight
        for message in messages:
            new_weight = getattr(message, "weight", None)
            if current_role != message.role or (new_weight != current_weight):
                aggregated_messages.extend(self._aggregate_role(messages_to_aggregate, current_role, latest_call_ids))

                if current_role == Roles.assistant:
                    assistant_message = aggregated_messages[-1]
                    assert isinstance(aggregated_messages[-1], AssistantMessage)
                    if assistant_message.tool_calls is not None:
                        for tool_call in assistant_message.tool_calls:
                            latest_call_ids.append(tool_call.id)

                elif current_role == Roles.tool:
                    # reordering only depends on the latest aggregated assistant message
                    latest_call_ids.clear()

                messages_to_aggregate.clear()
            current_weight = new_weight
            current_role = message.role
            messages_to_aggregate.append(message)

        # Add the last set of messages
        aggregated_messages.extend(self._aggregate_role(messages_to_aggregate, current_role, latest_call_ids))

        # If the first message is not a user message, or we didn't aggregate
        # anything (all system messages) for example, add an empty user message
        if len(aggregated_messages) == 0 or (
            not self._system_prompt_in_begin and aggregated_messages[0].role != Roles.user
        ):
            aggregated_messages.insert(0, self._user_message_class(content=""))

        return aggregated_messages

    def from_chat_completion_request(self, request: ChatCompletionRequest[UATS]) -> InstructRequestType:
        r"""Converts a chat completion request to an instruct request.

        Args:
            request: The chat completion request to convert.

        Returns:
            The converted instruct request.

        Examples:
            >>> from mistral_common.protocol.instruct.messages import UserMessage, AssistantMessage
            >>> request = ChatCompletionRequest(
            ...     messages=[
            ...         UserMessage(content="Hello"),
            ...         AssistantMessage(content="Hi"),
            ...     ],
            ... )
            >>> normalizer = InstructRequestNormalizer.normalizer()
            >>> instruct_request = normalizer.from_chat_completion_request(request)
        """
        system_prompt = self._aggregate_system_prompts(request.messages)
        messages = self._aggregate_messages(request.messages)

        settings = self.build_settings(request)
        if settings != ModelSettings.none():
            raise InvalidRequestException(f"Model settings are not supported for {type(self).__name__}, got {settings}")

        return self._instruct_request_class(
            messages=messages,
            system_prompt=system_prompt,
            available_tools=request.tools,
            continue_final_message=request.continue_final_message,
            settings=settings,
        )


class InstructRequestNormalizerV7(InstructRequestNormalizer):
    r"""Normalizer for the v7 tokenizer.

    Examples:
        >>> normalizer = InstructRequestNormalizerV7.normalizer()
    """

    _system_prompt_in_begin: bool = True
    _allow_tool_call_and_content: bool = True

    @staticmethod
    def normalizer(model_settings_builder: ModelSettingsBuilder | None = None) -> "InstructRequestNormalizerV7":
        r"""Returns a normalizer for the default instruct request.

        Args:
            model_settings_builder: Must be None for this normalizer version.

        Returns:
            A normalizer for the V7 instruct request.

        Raises:
            ValueError: If model_settings_builder is not None.

        Examples:
            >>> normalizer = InstructRequestNormalizerV7.normalizer()
        """
        if model_settings_builder is not None:
            raise ValueError(
                f"model_settings_builder must be None for InstructRequestNormalizerV7, got {model_settings_builder}"
            )
        return InstructRequestNormalizerV7(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest[UATS, Tool], None
        )

    def _aggregate_tool_messages(self, messages: list[UATS], latest_call_ids: list[str]) -> list[ToolMessageType]:
        """Normalize tool messages without JSON normalization.

        V7+ normalizers skip JSON content normalization for tool messages (chunk-type validation is
        handled by the validator).
        """
        tool_messages: list[ToolMessageType] = []
        for message in messages:
            assert isinstance(message, self._tool_message_class), "Expected tool message"
            content = self._aggregate_content_chunks([message])
            assert isinstance(content, str), (
                f"Unexpected content chunk types in tool message: {[type(c).__name__ for c in content]}"
            )
            tool_messages.append(
                self._tool_message_class(content=content, tool_call_id=message.tool_call_id, name=message.name)
            )
        return tool_messages

    def _aggregate_system_messages(self, messages: list[UATS]) -> list[SystemMessageType]:
        aggregated: list[SystemMessageType] = []
        for message in messages:
            if isinstance(message, self._system_message_class):
                content = self._aggregate_content_chunks([message])
                aggregated.append(self._system_message_class(content=content))
        return aggregated

    def _aggregate_role(self, messages: list[UATS], role: Roles | None, latest_call_ids: list[str]) -> Sequence[UATS]:
        if role == Roles.tool:
            return self._aggregate_tool_messages(messages, latest_call_ids)
        elif role == Roles.assistant:
            return [self._aggregate_assistant_messages(messages)]
        elif role == Roles.user:
            return [self._aggregate_user_messages(messages)]
        elif role == Roles.system:
            return self._aggregate_system_messages(messages)
        else:
            assert role is None and len(messages) == 0
            return []

    def _aggregate_system_prompts(self, messages: list[UATS]) -> str | None:
        raise NotImplementedError("We should not aggregate system prompts")

    def from_chat_completion_request(self, request: ChatCompletionRequest[UATS]) -> InstructRequestType:  # type: ignore[type-var, misc]
        r"""Converts a chat completion request to an instruct request.

        Args:
            request: The chat completion request to convert.

        Returns:
            The converted instruct request.

        Examples:
            >>> from mistral_common.protocol.instruct.messages import UserMessage, AssistantMessage
            >>> request = ChatCompletionRequest(
            ...     messages=[
            ...         UserMessage(content="Hello"),
            ...         AssistantMessage(content="Hi"),
            ...     ],
            ... )
            >>> normalizer = InstructRequestNormalizerV7.normalizer()
            >>> instruct_request = normalizer.from_chat_completion_request(request)
        """
        messages = self._aggregate_messages(request.messages)
        settings = self.build_settings(request)
        if settings != ModelSettings.none():
            raise InvalidRequestException(f"Model settings are not supported for {type(self).__name__}, got {settings}")
        return self._instruct_request_class(  # type: ignore[no-any-return]
            messages=messages,
            system_prompt=None,
            available_tools=request.tools,
            continue_final_message=request.continue_final_message,
            settings=settings,
        )


class InstructRequestNormalizerV13(InstructRequestNormalizerV7):
    r"""Normalizer for the v13 tokenizer.

    It reorders tool messages based on the tool call order.

    Examples:
        >>> normalizer = InstructRequestNormalizerV13.normalizer()
    """

    @staticmethod
    def normalizer(model_settings_builder: ModelSettingsBuilder | None = None) -> "InstructRequestNormalizerV13":
        r"""Returns a normalizer for the default instruct request.

        Args:
            model_settings_builder: Must be None for this normalizer version.

        Returns:
            A normalizer for the V13 instruct request.

        Raises:
            ValueError: If model_settings_builder is not None.
        """
        if model_settings_builder is not None:
            raise ValueError(
                f"model_settings_builder must be None for InstructRequestNormalizerV13, got {model_settings_builder}"
            )
        return InstructRequestNormalizerV13(
            UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest[UATS, Tool], None
        )

    @staticmethod
    def _inplace_sort_tool_messages(tool_messages: list[ToolMessageType], latest_call_ids: list[str]) -> None:
        id_to_tool_call_idx = {call_id: idx for idx, call_id in enumerate(latest_call_ids)}
        id_to_tool_result_idx = {message.tool_call_id: idx for idx, message in enumerate(tool_messages)}
        # First order by tool call idx and then by tool result idx
        tool_messages.sort(
            key=lambda msg: (
                id_to_tool_call_idx.get(msg.tool_call_id or "null", float("inf")),
                id_to_tool_result_idx[msg.tool_call_id],
            ),
        )

    def _aggregate_tool_messages(self, messages: list[UATS], latest_call_ids: list[str]) -> list[ToolMessageType]:
        tool_messages: list[ToolMessageType] = super()._aggregate_tool_messages(messages, latest_call_ids)
        self._inplace_sort_tool_messages(tool_messages=tool_messages, latest_call_ids=latest_call_ids)
        return tool_messages


class InstructRequestNormalizerV15(InstructRequestNormalizerV13):
    r"""Normalizer for the v15 tokenizer.

    It reorders tool messages based on the tool call order and builds model settings.

    Examples:
        >>> normalizer = InstructRequestNormalizerV15.normalizer()
    """

    _chunk_join_str: str = ""

    def _aggregate_tool_messages(self, messages: list[UATS], latest_call_ids: list[str]) -> list[ToolMessageType]:
        r"""V15 keeps all aggregated tool content (validation handled by the validator)."""
        tool_messages: list[ToolMessageType] = []
        for message in messages:
            assert isinstance(message, self._tool_message_class), "Expected tool message"
            content = self._aggregate_content_chunks([message])
            tool_messages.append(
                self._tool_message_class(content=content, tool_call_id=message.tool_call_id, name=message.name)
            )
        self._inplace_sort_tool_messages(tool_messages=tool_messages, latest_call_ids=latest_call_ids)
        return tool_messages

    @staticmethod
    def normalizer(model_settings_builder: ModelSettingsBuilder | None = None) -> "InstructRequestNormalizerV15":
        r"""Returns a normalizer for the V15 instruct request.

        Args:
            model_settings_builder: The builder for model settings.

        Returns:
            A normalizer for the V15 instruct request.
        """
        return InstructRequestNormalizerV15(
            UserMessage,
            AssistantMessage,
            ToolMessage,
            SystemMessage,
            InstructRequest,
            model_settings_builder,
        )

    def build_settings(self, request: ChatCompletionRequest) -> ModelSettings:
        r"""Build model settings using the configured model settings builder.

        Args:
            request: The chat completion request.

        Returns:
            The built model settings.

        Raises:
            InvalidRequestException: If no model settings builder is configured.
        """
        if self._model_settings_builder is None:
            raise InvalidRequestException(f"model_settings_builder must not be None for {type(self).__name__}")
        return self._model_settings_builder.build_settings(request)

    def from_chat_completion_request(self, request: ChatCompletionRequest[UATS]) -> InstructRequestType:  # type: ignore[type-var, misc]
        r"""Converts a chat completion request to an instruct request.

        Args:
            request: The chat completion request to convert.

        Returns:
            The converted instruct request.
        """
        messages = self._aggregate_messages(request.messages)
        settings = self.build_settings(request)
        return self._instruct_request_class(  # type: ignore[no-any-return]
            messages=messages,
            system_prompt=None,
            available_tools=request.tools,
            continue_final_message=request.continue_final_message,
            settings=settings,
        )


def normalizer_for_tokenizer_version(
    version: TokenizerVersion, model_settings_builder: ModelSettingsBuilder | None = None
) -> InstructRequestNormalizer:
    r"""Deprecated in favor to `get_normalizer`, will be removed in 1.12.0."""
    warnings.warn(
        "`normalizer_for_tokenizer_version` is deprecated and will be removed in 1.12.0. "
        "Please call `get_normalizer` instead.",
        FutureWarning,
    )
    return get_normalizer(version=version, model_settings_builder=model_settings_builder)


def get_normalizer(
    version: TokenizerVersion, model_settings_builder: ModelSettingsBuilder | None = None
) -> InstructRequestNormalizer:
    r"""Gets the appropriate normalizer for the given tokenizer version.

    Args:
        version: The tokenizer version to get the normalizer for.
        model_settings_builder: The builder for model settings, or None if unsupported.

    Returns:
        The appropriate normalizer for the given tokenizer version.

    Examples:
        >>> normalizer = get_normalizer(TokenizerVersion.v1)
    """
    match version:
        case TokenizerVersion.v1 | TokenizerVersion.v2 | TokenizerVersion.v3:
            normalizer_cls = InstructRequestNormalizer
        case TokenizerVersion.v7:
            normalizer_cls = InstructRequestNormalizerV7
        case TokenizerVersion.v11 | TokenizerVersion.v13:
            normalizer_cls = InstructRequestNormalizerV13
        case TokenizerVersion.v15:
            normalizer_cls = InstructRequestNormalizerV15
        case _:
            assert_never(version)

    return normalizer_cls.normalizer(model_settings_builder=model_settings_builder)



import re
from unittest.mock import MagicMock

import llguidance as llg
import pytest

from mistral_common.guidance.tokenizer import MistralLLGTokenizer, from_mistral_tokenizer
from mistral_common.protocol.instruct.normalize import get_normalizer
from mistral_common.protocol.instruct.validator import ValidationMode, get_validator
from mistral_common.tokens.tokenizers.base import SpecialTokenPolicy, SpecialTokens, Tokenizer, TokenizerVersion
from mistral_common.tokens.tokenizers.instruct import InstructTokenizerV7
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.tokens.tokenizers.tekken import SpecialTokenInfo, Tekkenizer
from tests.test_tekken import get_special_tokens, quick_vocab

_NUM_SPECIAL_TOKENS = 100
_EXTRA_TOKENS = [b"a", b"b", b"c", b"f", b"de", b"he", b"llo"]


@pytest.fixture(scope="module")
def tekkenizer() -> Tekkenizer:
    special_tokens = get_special_tokens(TokenizerVersion.v7)
    return Tekkenizer(
        quick_vocab(_EXTRA_TOKENS),
        special_tokens=special_tokens,
        pattern=r".+",
        vocab_size=256 + _NUM_SPECIAL_TOKENS,
        num_special_tokens=_NUM_SPECIAL_TOKENS,
        version=TokenizerVersion.v7,
    )


@pytest.fixture(scope="module")
def llg_tokenizer(tekkenizer: Tekkenizer) -> MistralLLGTokenizer:
    return MistralLLGTokenizer(tekkenizer)


@pytest.fixture(scope="module")
def mistral_tokenizer(tekkenizer: Tekkenizer) -> MistralTokenizer:
    instruct_tokenizer = InstructTokenizerV7(tekkenizer)
    normalizer = get_normalizer(tekkenizer.version, tekkenizer.model_settings_builder)
    validator = get_validator(tekkenizer.version, mode=ValidationMode.test)
    return MistralTokenizer(instruct_tokenizer, validator=validator, request_normalizer=normalizer)


@pytest.fixture(scope="module")
def ll_tokenizer(mistral_tokenizer: MistralTokenizer) -> llg.LLTokenizer:
    return from_mistral_tokenizer(mistral_tokenizer)


class TestMistralLLGTokenizer:
    def test_init_rejects_non_tekkenizer(self) -> None:
        mock_tokenizer = MagicMock(spec=Tokenizer)
        with pytest.raises(TypeError, match="Guidance only supports Tekken tokenizers"):
            MistralLLGTokenizer(mock_tokenizer)

    def test_eos_and_bos_ids(self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer) -> None:
        assert llg_tokenizer.eos_token_id == tekkenizer.eos_id
        assert llg_tokenizer.bos_token_id == tekkenizer.bos_id

    def test_tokens_length(self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer) -> None:
        assert len(llg_tokenizer.tokens) == tekkenizer.n_words

    def test_special_token_ids_count(self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer) -> None:
        assert len(llg_tokenizer.special_token_ids) == tekkenizer.num_special_tokens

    def test_all_special_tokens_are_angle_bracketed(self, llg_tokenizer: MistralLLGTokenizer) -> None:
        for i in llg_tokenizer.special_token_ids:
            token_bytes = llg_tokenizer.tokens[i]
            token_str = token_bytes.decode("utf-8")
            assert re.fullmatch(r"<.*>", token_str), f"Special token at id={i} is not angle-bracketed: {token_str!r}"

    def test_special_token_conversion(self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer) -> None:
        checked = 0
        for token in SpecialTokens:
            try:
                rank = tekkenizer.get_special_token(token.value)
            except ValueError:
                continue

            llg_bytes = llg_tokenizer.tokens[rank]
            if re.fullmatch(r"\[.*\]", token.value):
                expected = token.value.replace("[", "<").replace("]", ">")
            else:
                expected = token.value

            assert llg_bytes == expected.encode("utf-8"), (
                f"Token {token.name} at rank={rank}: expected {expected!r}, got {llg_bytes!r}"
            )
            checked += 1

        # Filler tokens (<SPECIAL_XX>) should be preserved as-is
        for i in range(tekkenizer.num_special_tokens):
            piece = tekkenizer.id_to_piece(i)
            if re.fullmatch(r"<SPECIAL_\d+>", piece):
                assert llg_tokenizer.tokens[i] == piece.encode("utf-8"), (
                    f"Filler token at id={i}: expected {piece!r}, got {llg_tokenizer.tokens[i]!r}"
                )
                checked += 1

        assert checked == tekkenizer.num_special_tokens

    def test_non_special_tokens_match_byte_pieces(
        self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer
    ) -> None:
        for i in range(tekkenizer.num_special_tokens, tekkenizer.n_words):
            expected = tekkenizer.id_to_byte_piece(i, SpecialTokenPolicy.RAISE)
            assert llg_tokenizer.tokens[i] == expected, (
                f"Token at id={i}: expected {expected!r}, got {llg_tokenizer.tokens[i]!r}"
            )

    def test_call_encodes_string(self, tekkenizer: Tekkenizer, llg_tokenizer: MistralLLGTokenizer) -> None:
        test_string = "abc"
        assert llg_tokenizer(test_string) == tekkenizer.encode(test_string, bos=False, eos=False)

    def test_init_rejects_invalid_special_token_format(self) -> None:
        base = list(Tekkenizer.DEPRECATED_SPECIAL_TOKENS)
        next_rank = len(base)
        special_tokens: list[SpecialTokenInfo] = [
            *base,
            SpecialTokenInfo(rank=next_rank, token_str="INVALID_NO_BRACKETS", is_control=True),
        ]
        vocab = quick_vocab()
        num_special = len(special_tokens)
        tekkenizer = Tekkenizer(
            vocab,
            special_tokens=special_tokens,
            pattern=r".+",
            vocab_size=len(vocab) + num_special,
            num_special_tokens=num_special,
            version=TokenizerVersion.v7,
        )
        with pytest.raises(ValueError, match="Invalid special token"):
            MistralLLGTokenizer(tekkenizer)

    def test_init_rejects_duplicate_special_tokens(self) -> None:
        base = list(Tekkenizer.DEPRECATED_SPECIAL_TOKENS)
        next_rank = len(base)
        # Both [CUSTOM] and <CUSTOM> map to <CUSTOM> after bracket conversion
        special_tokens: list[SpecialTokenInfo] = [
            *base,
            SpecialTokenInfo(rank=next_rank, token_str="<CUSTOM>", is_control=True),
            SpecialTokenInfo(rank=next_rank + 1, token_str="[CUSTOM]", is_control=True),
        ]
        vocab = quick_vocab()
        num_special = len(special_tokens)
        tekkenizer = Tekkenizer(
            vocab,
            special_tokens=special_tokens,
            pattern=r".+",
            vocab_size=len(vocab) + num_special,
            num_special_tokens=num_special,
            version=TokenizerVersion.v7,
        )
        with pytest.raises(ValueError, match="Duplicate special token"):
            MistralLLGTokenizer(tekkenizer)


class TestFromMistralTokenizer:
    def test_properties(self, tekkenizer: Tekkenizer, ll_tokenizer: llg.LLTokenizer) -> None:
        assert isinstance(ll_tokenizer, llg.LLTokenizer)
        assert ll_tokenizer.vocab_size == tekkenizer.n_words
        assert ll_tokenizer.eos_token == tekkenizer.eos_id
        for i in range(tekkenizer.num_special_tokens):
            assert ll_tokenizer.is_special_token(i), f"Token id={i} should be special"

    def test_tokenize_str_matches_tekkenizer(self, tekkenizer: Tekkenizer, ll_tokenizer: llg.LLTokenizer) -> None:
        test_strings = ["abc", "hello", "de", "abcdefhello"]
        for s in test_strings:
            expected = tekkenizer.encode(s, bos=False, eos=False)
            result = ll_tokenizer.tokenize_str(s)
            assert result == expected, f"Mismatch for {s!r}: expected {expected}, got {result}"

    def test_decode_str_roundtrip(self, tekkenizer: Tekkenizer, ll_tokenizer: llg.LLTokenizer) -> None:
        test_strings = ["abc", "hello", "de"]
        for s in test_strings:
            tokens = tekkenizer.encode(s, bos=False, eos=False)
            decoded = ll_tokenizer.decode_str(tokens)
            assert decoded == s, f"Roundtrip failed for {s!r}: got {decoded!r}"

    def test_decode_bytes_roundtrip(self, tekkenizer: Tekkenizer, ll_tokenizer: llg.LLTokenizer) -> None:
        test_strings = ["abc", "hello"]
        for s in test_strings:
            tokens = tekkenizer.encode(s, bos=False, eos=False)
            decoded = ll_tokenizer.decode_bytes(tokens)
            assert decoded == s.encode("utf-8"), f"Roundtrip failed for {s!r}: got {decoded!r}"

    def test_decode_special_tokens(self, tekkenizer: Tekkenizer, ll_tokenizer: llg.LLTokenizer) -> None:
        for token in SpecialTokens:
            try:
                rank = tekkenizer.get_special_token(token.value)
            except ValueError:
                continue

            if re.fullmatch(r"\[.*\]", token.value):
                expected = token.value.replace("[", "<").replace("]", ">")
            else:
                expected = token.value

            decoded_str = ll_tokenizer.decode_str([rank])
            assert decoded_str == expected, (
                f"decode_str for {token.name} (id={rank}): expected {expected!r}, got {decoded_str!r}"
            )
            decoded_bytes = ll_tokenizer.decode_bytes([rank])
            assert decoded_bytes == expected.encode("utf-8"), (
                f"decode_bytes for {token.name} (id={rank}): "
                f"expected {expected.encode('utf-8')!r}, got {decoded_bytes!r}"
            )



import pytest

from mistral_common.exceptions import (
    InvalidAssistantMessageException,
    InvalidFunctionCallException,
    InvalidToolMessageException,
)
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    Roles,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.tool_calls import FunctionCall, ToolCall
from mistral_common.protocol.instruct.validator import (
    MistralRequestValidator,
    MistralRequestValidatorV3,
    ValidationMode,
)


@pytest.fixture(
    params=[
        MistralRequestValidator(ValidationMode.serving),
        MistralRequestValidatorV3(ValidationMode.serving),
    ]
)
def validator(request: pytest.FixtureRequest) -> MistralRequestValidator:
    return request.param  # type: ignore


class TestValidateAssistantMessage:
    def test_content_and_tool_calls(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(
            InvalidAssistantMessageException,
            match=r"Assistant message must have either content or tool_calls, but not both.",
        ):
            validator._validate_assistant_message(
                AssistantMessage(
                    role=Roles.assistant,
                    content="a",
                    tool_calls=[ToolCall(function=FunctionCall(name="a", arguments='{"a": 1}'))],
                ),
            )

    @pytest.mark.parametrize("content,tool_calls", [("", None), (None, None), ("", []), (None, [])])
    def test_no_content_and_no_tool_calls(
        self, content: str | None, tool_calls: list | None, validator: MistralRequestValidator
    ) -> None:
        with pytest.raises(
            InvalidAssistantMessageException,
            match=r"Assistant message must have either content or tool_calls, but not both.",
        ):
            validator._validate_assistant_message(
                AssistantMessage(role=Roles.assistant, content=content, tool_calls=tool_calls),
            )

    def test_no_content_and_tool_calls(self, validator: MistralRequestValidator) -> None:
        validator._validate_assistant_message(
            AssistantMessage(
                role=Roles.assistant,
                content="",
                tool_calls=[
                    ToolCall(
                        id="AbCd56789",
                        function=FunctionCall(name="function_call", arguments='{"a": 1}'),
                    )
                ],
            ),
        )

    def test_content_and_no_tool_calls(self, validator: MistralRequestValidator) -> None:
        validator._validate_assistant_message(
            AssistantMessage(
                role=Roles.assistant,
                content="content",
                tool_calls=[],
            ),
        )

    def test_bad_function_name(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidFunctionCallException, match=r"must be a-z, A-Z, 0-9"):
            validator._validate_assistant_message(
                AssistantMessage(
                    role=Roles.assistant,
                    tool_calls=[
                        ToolCall(
                            id="AbCd56789",
                            function=FunctionCall(name=')_(_)~@:}>""', arguments='{"a": 1}'),
                        )
                    ],
                ),
            )

    @pytest.mark.skip(reason="If json parse fails we assume that its plain text like '22' or 'a'")
    def test_invalid_function_json(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidFunctionCallException, match=r"must be a-z, A-Z, 0-9"):
            validator._validate_assistant_message(
                AssistantMessage(
                    role=Roles.assistant,
                    tool_calls=[ToolCall(function=FunctionCall(name="function_name", arguments='{"a":aewfawef 1}'))],
                )
            )

    def test_tool_id_defined(self, validator: MistralRequestValidator) -> None:
        if not isinstance(validator, MistralRequestValidatorV3):
            pytest.skip("MistralRequestValidator v1 does not validate tool call ids")
        with pytest.raises(InvalidFunctionCallException, match=r"Tool call id has to be defined"):
            validator._validate_assistant_message(
                AssistantMessage(
                    role=Roles.assistant,
                    tool_calls=[ToolCall(function=FunctionCall(name="foo", arguments='{"a": 1}'))],
                )
            )

    def test_tool_calls_bad_id(self, validator: MistralRequestValidator) -> None:
        if not isinstance(validator, MistralRequestValidatorV3):
            pytest.skip("MistralRequestValidator v1 does not validate tool call ids")
        with pytest.raises(
            InvalidFunctionCallException,
            match=r"Tool call id was bad_id but must be a-z, A-Z, 0-9, with a length of 9.",
        ):
            validator._validate_assistant_message(
                AssistantMessage(
                    tool_calls=[
                        ToolCall(
                            id="bad_id",
                            function=FunctionCall(name="foo", arguments="{}"),
                        )
                    ]
                )
            )


class TestValidateUserMessage:
    def test_user_message_blank_content_OK(self, validator: MistralRequestValidator) -> None:
        validator._validate_user_message(UserMessage(content=""))


class TestValidateToolMessage:
    def test_tool_message_bad_name(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidToolMessageException, match=r"must be a-z, A-Z, 0-9"):
            validator._validate_tool_message(ToolMessage(name="~@}£Q_+", content="{}"))

    @pytest.mark.skip(reason="If json parse fails we assume that its plain text like '22' or 'a'")
    def test_tool_message_invalid_json(self, validator: MistralRequestValidator) -> None:
        with pytest.raises(InvalidToolMessageException, match=r"must be a-z, A-Z, 0-9"):
            validator._validate_tool_message(ToolMessage(name="function_name", content='{"afewfaef}'))

    def test_tool_message_bad_id(self, validator: MistralRequestValidator) -> None:
        if not isinstance(validator, MistralRequestValidatorV3):
            pytest.skip("MistralRequestValidator v1 does not validate tool call ids")
        with pytest.raises(
            InvalidToolMessageException,
            match=r"Tool call id was bad_id but must be a-z, A-Z, 0-9, with a length of 9.",
        ):
            validator._validate_tool_message(
                ToolMessage(name="foo", content="bar", tool_call_id="bad_id"),
            )



r"""Reusable helper functions for chat template tests.

This module contains utility functions used by test files across the
`unit/` and `transformers/` subdirectories.  It intentionally has no
fixtures or test data -- those live in `conftest.py` and `fixtures_data.py`.

HuggingFace-specific helpers live in `hf_utils.py`.
"""

import json
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
from jinja2 import BaseLoader
from jinja2.sandbox import ImmutableSandboxedEnvironment
from PIL import Image

from mistral_common.integrations.chat_templates.template_generator import TemplateConfig
from mistral_common.protocol.instruct.request import ChatCompletionRequest, ReasoningEffort
from mistral_common.protocol.instruct.validator import ValidationMode
from mistral_common.tokens.tokenizers.audio import Audio
from mistral_common.tokens.tokenizers.base import TokenizerVersion
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.tokens.tokenizers.model_settings_builder import EnumBuilder, ModelSettingsBuilder
from tests.test_tekken import get_special_tokens

# Golden template files live in the data/ tree (outside src/).
_GOLDEN_DIR = Path(__file__).parent.parent.parent / "data" / "chat_templates"

SPM_WHITESPACE = "▁"


@dataclass(frozen=True)
class TestConfig:
    r"""Test configuration for chat template parametrization."""

    __test__ = False

    version: TokenizerVersion
    spm: bool = False
    image: bool = False
    audio: bool = False
    think: bool = False
    plain_think: bool = False


def _make_config(c: TestConfig) -> TemplateConfig:
    r"""Create a `TemplateConfig` from a test config."""
    return TemplateConfig(
        version=c.version,
        spm=c.spm,
        image_support=c.image,
        audio_support=c.audio,
        thinking_support=c.think,
        plain_thinking_support=c.plain_think,
        use_special_token_variables=True,
    )


def _load_golden_template(config: TemplateConfig) -> str:
    r"""Load the static golden template for a config."""
    parts = [config.version.value]
    if config.image_support and config.any_thinking_support:
        parts.append("image_think")
    elif config.image_support:
        parts.append("image")
    elif config.audio_support:
        parts.append("audio")
    elif config.any_thinking_support:
        parts.append("think")
    if config.spm:
        parts.append("spm")
    filename = "_".join(parts) + ".jinja"
    path = _GOLDEN_DIR / filename
    if not path.exists():
        raise FileNotFoundError(f"Golden template not found: {path}")
    return path.read_text()


def render_template(
    template: str, messages: list[Any], tools: list[Any] | None = None, reasoning_effort: str | None = None
) -> str:
    r"""Render a Jinja2 template with the given messages using a pure Jinja2 sandbox.

    This function tests template rendering independently of HuggingFace transformers.
    It uses ImmutableSandboxedEnvironment directly, while `encode_transformers`
    tests integration with the HuggingFace `render_jinja_template` function.
    Both should produce identical output for the same input.
    """

    def raise_exception(msg: str) -> None:
        raise ValueError(msg)

    env = ImmutableSandboxedEnvironment(loader=BaseLoader())
    env.globals["raise_exception"] = raise_exception
    jinja_template = env.from_string(template)

    render_kwargs: dict[str, Any] = {
        "messages": messages,
        "tools": tools,
        "bos_token": "<s>",
        "eos_token": "</s>",
    }

    # Only add reasoning_effort for v15+ templates that support it
    if reasoning_effort is not None:
        render_kwargs["reasoning_effort"] = reasoning_effort

    return jinja_template.render(**render_kwargs)


def encode_mistral_common(mistral_tokenizer: MistralTokenizer, chat_request: ChatCompletionRequest, spm: bool) -> str:
    r"""Encode a chat request using mistral-common tokenizer.

    Returns the text representation (not token IDs) for parity comparison
    with the Jinja template rendering. Token ID comparison is handled
    separately via `encode_hf_tokens`.
    """
    mistral_encoded = str(mistral_tokenizer.encode_chat_completion(chat_request).text)
    # Collapse each image token sequence ([IMG]...[IMG_BREAK]...[IMG_END]) into a single [IMG].
    # Each sequence has exactly one [IMG_END], so replacing inner tokens and converting [IMG_END]
    # preserves one [IMG] per image (important for multi-image inputs).
    mistral_encoded = mistral_encoded.replace("[IMG]", "").replace("[IMG_BREAK]", "").replace("[IMG_END]", "[IMG]")
    # Remove audio tokens except one per audio
    mistral_encoded = mistral_encoded.replace("[AUDIO]", "").replace("[BEGIN_AUDIO]", "[AUDIO]")
    if spm:
        mistral_encoded = mistral_encoded.replace(SPM_WHITESPACE, " ").replace("<0x0A>", "\n")
    return mistral_encoded


def _create_dummy_image() -> Image.Image:
    r"""Create a simple dummy 2x2 red square image for testing."""
    return Image.new("RGB", (2, 2), color="red")


def _sin_wave(sampling_rate: int, duration: float) -> np.ndarray:
    r"""Generate a sine wave numpy array."""
    return np.sin(np.ones([int(duration * sampling_rate)]))


def _sample_audio() -> Audio:
    r"""Create a sample `Audio` instance for testing."""
    sampling_rate = 44100
    original_array = _sin_wave(sampling_rate, 1)
    return Audio(
        audio_array=original_array,
        sampling_rate=sampling_rate,
        format="wav",
    )


_IMAGE_URL = "https://upload.wikimedia.org/wikipedia/commons/7/78/Red_Square_%282x2_Pixel%29.png"
_IMAGE = _create_dummy_image()
_AUDIO_URL = _sample_audio().to_base64("wav")
_AUDIO = _AUDIO_URL


def _build_tekken_json(config: TestConfig, output_dir: Path) -> Path:
    r"""Build a tekken.json file for the given test config.

    Constructs a complete Tekken tokenizer JSON file by combining the base
    vocabulary from the shipped tokenizer with version-specific special tokens
    and optional image/audio/model_settings configuration. The file is written
    to `output_dir / "tekken.json"` and can be loaded by
    `MistralTokenizer.from_file`.

    Args:
        config: Test configuration specifying version and feature flags.
        output_dir: Directory to write the JSON file into.

    Returns:
        Path to the written tekken.json file.
    """
    with open(MistralTokenizer._data_path() / "tekken_240911.json", "r", encoding="utf-8") as f:
        base_data = json.load(f)

    special_tokens = get_special_tokens(
        tokenizer_version=config.version, add_audio=config.audio, add_think=config.think
    )

    tekken_data: dict[str, Any] = {
        "config": {
            "pattern": base_data["config"]["pattern"],
            "default_vocab_size": base_data["config"]["default_vocab_size"],
            "default_num_special_tokens": 100,
            "version": config.version.value,
        },
        "vocab": base_data["vocab"],
        "special_tokens": special_tokens,
    }

    if config.image:
        tekken_data["image"] = {
            "image_patch_size": 2,
            "max_image_size": 10,
            "spatial_merge_size": 1,
        }

    if config.audio:
        tekken_data["audio"] = {
            "sampling_rate": 24_000,
            "frame_rate": 12.5,
            "audio_encoding_config": {
                "num_mel_bins": 128,
                "window_size": 400,
                "hop_length": 160,
            },
        }

    if config.version.supports_model_settings:
        model_settings_builder = ModelSettingsBuilder(
            reasoning_effort=EnumBuilder(
                accepts_none=True,
                default=ReasoningEffort.none,
                values=[ReasoningEffort.none, ReasoningEffort.high],
            )
        )
        tekken_data["model_settings_builder"] = model_settings_builder.model_dump(mode="json")

    tekken_path = output_dir / "tekken.json"
    with open(tekken_path, "w", encoding="utf-8") as f:
        json.dump(tekken_data, f, ensure_ascii=False)

    return tekken_path


def _build_spm_path(config: TestConfig, output_dir: Path) -> Path:
    r"""Copy the SPM model file with the correct version suffix.

    The SPM tokenizer version is determined by the filename suffix (e.g.,
    `.model.v3` for v3). Image support is indicated by an `m1` suffix
    (e.g., `.model.v3m1`). The shipped v7m1 model file is copied with the
    appropriate suffix for the requested config.

    Args:
        config: Test configuration specifying version and image flag.
        output_dir: Directory to copy the file into.

    Returns:
        Path to the copied model file.
    """
    source = MistralTokenizer._data_path() / "mistral_instruct_tokenizer_241114.model.v7m1"
    suffix = f".model.{config.version.value}"
    if config.image:
        suffix += "m1"
    dest = output_dir / f"tokenizer{suffix}"
    shutil.copy2(source, dest)
    return dest


def _get_mistral_tekkenizer(config: TestConfig, output_dir: Path, validation_mode: ValidationMode) -> MistralTokenizer:
    r"""Build a `MistralTokenizer` with Tekken backend via `from_file`.

    Writes a tekken.json with the desired version and features via
    `_build_tekken_json`, then loads it through the production
    `MistralTokenizer.from_file` path.

    Args:
        config: Test configuration specifying version, image, audio, think.
        output_dir: Directory to write the temporary tekken.json file.
        validation_mode: Validation mode (test or finetuning).

    Returns:
        A configured `MistralTokenizer` instance.
    """
    tekken_path = _build_tekken_json(config, output_dir)
    return MistralTokenizer.from_file(str(tekken_path), mode=validation_mode)


def _get_mistral_sentencepiece(
    config: TestConfig, output_dir: Path, validation_mode: ValidationMode
) -> MistralTokenizer:
    r"""Build a `MistralTokenizer` with SentencePiece backend via `from_file`.

    Copies the shipped SPM model with the correct version suffix via
    `_build_spm_path`, then loads it through the production
    `MistralTokenizer.from_file` path.

    Args:
        config: Test configuration specifying version and image flag.
        output_dir: Directory to copy the SPM model file into.
        validation_mode: Validation mode (test or finetuning).

    Returns:
        A configured `MistralTokenizer` instance.
    """
    spm_path = _build_spm_path(config, output_dir)
    return MistralTokenizer.from_file(str(spm_path), mode=validation_mode)


def _get_mistral_tokenizer(
    spm: bool,
    tokenizer_version: TokenizerVersion,
    validation_mode: ValidationMode,
    image: bool,
    audio: bool,
    think: bool,
    output_dir: Path,
) -> MistralTokenizer:
    r"""Get a `MistralTokenizer` instance for testing.

    Args:
        spm: Whether to use SentencePiece backend instead of Tekken.
        tokenizer_version: Version of the tokenizer to build.
        validation_mode: Validation mode (test or finetuning).
        image: Whether to attach an image encoder.
        audio: Whether to attach an audio encoder (Tekken only).
        think: Whether to enable thinking special tokens (Tekken only).
        output_dir: Temporary directory for tokenizer files.

    Returns:
        A configured `MistralTokenizer` instance.
    """
    config = TestConfig(
        version=tokenizer_version,
        spm=spm,
        image=image,
        audio=audio,
        think=think,
        plain_think=False,
    )
    if spm:
        return _get_mistral_sentencepiece(config, output_dir, validation_mode)

    return _get_mistral_tekkenizer(config, output_dir, validation_mode)
